# # -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from dataclassy import dataclass
from loguru import logger as log

# user defined formula
from streaming_helper.utilities import string_modification as str_mod


def positions_and_orders(
    current_size: int,
    current_orders_size: int,
) -> int:
    """ """

    return current_size + current_orders_size


def ensure_sign_consistency(side) -> float:
    """ """
    return -1 if side == "sell" else 1


def proforma_size(
    current_size: int,
    current_orders_size: int,
    next_orders_size: int,
) -> int:
    """ """

    return (
        positions_and_orders(current_size, current_orders_size)
        + next_orders_size  # the sign is +
    )


def are_size_and_order_appropriate(
    purpose: str,
    current_size_or_open_position: float,
    current_orders_size: int,
    next_orders_size: int,
    max_position: float = None,
) -> bool:
    """
    purpose: add_position/reduce_position
    for reduce: open position/individual trade
    for add: current position
    """

    proforma = proforma_size(
        current_size_or_open_position, current_orders_size, next_orders_size
    )

    ordering_is_ok = False

    if purpose == "add_position":

        if max_position < 0:
            ordering_is_ok = (proforma) > (max_position)

        if max_position > 0:
            ordering_is_ok = (proforma) < (max_position)

    if purpose == "reduce_position":

        if current_size_or_open_position < 0:
            ordering_is_ok = proforma > current_size_or_open_position

        if current_size_or_open_position > 0:
            ordering_is_ok = proforma < current_size_or_open_position

    log.debug(
        f"ordering_is_ok  {ordering_is_ok} current_size_or_open_position \
        {current_size_or_open_position} proforma  {proforma} max_position  \
            {max_position} current_orders_size  {current_orders_size} next_orders_size  {next_orders_size} "
    )

    return ordering_is_ok


def check_if_next_closing_size_will_not_exceed_the_original(
    basic_size: int,
    net_size: int,
    next_size: int,
) -> bool:
    """ """

    basic_size_higher_than_next_closing_size = abs(basic_size) >= abs(next_size)
    basic_size_higher_than_net_size = abs(basic_size) >= abs(net_size)
    basic_size_plus_next_size = (next_size) + (basic_size)

    if abs(net_size) != abs(basic_size):
        pass

    if basic_size > 0:
        basic_size_sign_diff_than_next_size = basic_size_plus_next_size < basic_size

    if basic_size < 0:
        basic_size_sign_diff_than_next_size = basic_size_plus_next_size > basic_size

    # log.debug (f"basic_size_higher_than_next_closing_size {basic_size_higher_than_next_closing_size} basic_size_higher_than_net_size {basic_size_higher_than_net_size} basic_size_plus_next_size {basic_size_plus_next_size} basic_size_sign_diff_than_next_size {basic_size_sign_diff_than_next_size}")

    return (
        basic_size_higher_than_next_closing_size
        and basic_size_higher_than_net_size
        and basic_size_sign_diff_than_next_size
    )  # and net_size_exceeding_the_basic_size


def provide_size_to_close_transaction(
    basic_size: int,
    net_size: int,
) -> int:
    """ """

    next_size = min(basic_size, net_size)

    return abs(next_size)


def size_rounding(
    instrument_name: str,
    futures_instruments,
    proposed_size: float,
) -> int:
    """ """

    min_trade_amount = [
        o["min_trade_amount"]
        for o in futures_instruments
        if o["instrument_name"] == instrument_name
    ][0]

    rounded_size = round(proposed_size / min_trade_amount) * min_trade_amount

    return max(min_trade_amount, rounded_size)  # size is never 0


def delta(
    last_price: float,
    prev_price: float,
) -> float:
    """ """
    return last_price - prev_price


def delta_pct(
    last_price: float,
    prev_price: float,
) -> float:
    """ """
    return abs(delta(last_price, prev_price) / prev_price)


def get_label(
    status: str,
    label_main_or_label_transactions: str,
) -> str:
    """
    provide transaction label
    """

    if status == "open":

        # get open label
        label = str_mod.labelling("open", label_main_or_label_transactions)

    else:

        # parsing label id
        label_id: int = str_mod.parsing_label(label_main_or_label_transactions)["int"]

        # parsing label strategy
        label_main: str = str_mod.parsing_label(label_main_or_label_transactions)[
            "main"
        ]

        if status == "contra":

            if "closed" in label_main_or_label_transactions:
                label = f"""{label_main}-open-{label_id}"""

            if "open" in label_main_or_label_transactions:
                label = f"""{label_main}-closed-{label_id}"""

        if status == "closed":

            # combine id + label strategy
            label: str = f"""{label_main}-closed-{label_id}"""

    return label


def compute_profit_usd(
    transaction_price: float,
    currrent_price: float,
    size: int,
    side: str,
) -> float:
    """
    get difference between now and transaction time
    """
    if side == "buy":
        delta = currrent_price - transaction_price

    if side == "sell":
        delta = transaction_price - currrent_price

    profit = delta / transaction_price * abs(size)

    return profit


def pct_price_in_usd(
    price: float,
    pct_threshold: float,
) -> float:

    return price * pct_threshold


def price_plus_pct(
    price: float,
    pct_threshold: float,
) -> float:

    return price + pct_price_in_usd(
        price,
        pct_threshold,
    )


def price_minus_pct(
    price: float,
    pct_threshold: float,
) -> float:

    return price - pct_price_in_usd(
        price,
        pct_threshold,
    )


def is_transaction_price_minus_below_threshold(
    last_transaction_price: float,
    current_price: float,
    pct_threshold: float,
) -> bool:

    return price_minus_pct(last_transaction_price, pct_threshold) >= current_price


def is_transaction_price_plus_above_threshold(
    last_transaction_price: float,
    current_price: float,
    pct_threshold: float,
) -> bool:

    return (
        price_plus_pct(
            last_transaction_price,
            pct_threshold,
        )
        < current_price
    )


def profit_usd_has_exceed_target(
    target_profit: float,
    transaction_price: float,
    currrent_price: float,
    size: int,
    side: str,
) -> bool:
    """
    get difference between now and transaction time
    """
    current_profit = compute_profit_usd(
        transaction_price,
        currrent_price,
        size,
        side,
    )

    target = abs(pct_price_in_usd(target_profit, size))

    return current_profit > 0 and current_profit > target


def delta_time(
    server_time: int,
    time_stamp: int,
) -> int:
    """
    get difference between now and transaction time
    """

    return server_time - time_stamp


def is_minimum_waiting_time_has_passed(
    server_time: int,
    time_stamp: int,
    time_threshold: float,
) -> bool:
    """
    check whether delta time has exceed time threhold
    """

    return (
        True
        if time_stamp == []
        else delta_time(server_time, time_stamp) > time_threshold
    )


def get_max_time_stamp(result_strategy_label: list) -> int:
    """ """
    return (
        []
        if result_strategy_label == []
        else max([o["timestamp"] for o in result_strategy_label])
    )


def get_order_id_max_time_stamp(result_strategy_label: list) -> int:
    """ """
    return (
        0
        if get_max_time_stamp(result_strategy_label) == []
        else [
            o["order_id"]
            for o in result_strategy_label
            if o["timestamp"] == get_max_time_stamp(result_strategy_label)
        ][0]
    )


def get_transactions_len(result_strategy_label: list) -> int:
    """ """
    return 0 if result_strategy_label == [] else len([o for o in result_strategy_label])


def get_transactions_sum(result_strategy_label: list) -> int:
    """
    summing transaction under SAME strategy label
    """
    return (
        0
        if result_strategy_label == []
        else sum([o["amount"] for o in result_strategy_label])
    )


def get_transaction_side(transaction: dict) -> str:
    """ """

    try:
        return transaction["direction"]

    except:
        return transaction["side"]


def get_transaction_size(transaction: dict) -> int:
    """ """
    return transaction["amount"]


def get_transaction_instrument(transaction: dict) -> int:
    """ """
    return transaction["instrument_name"]


def get_transaction_label(transaction: dict) -> str:
    """ """
    return transaction["label"]


def get_transaction_price(transaction: dict) -> float:
    """ """
    return transaction["price"]


def get_label_integer(label: dict) -> bool:
    """ """

    return str_mod.parsing_label(label)["int"]


def get_order_label(data_from_db: list) -> list:
    """ """

    return [o["label"] for o in data_from_db]


def get_label_super_main(result: list, strategy_label: str) -> list:
    """ """

    return [
        o
        for o in result
        if str_mod.parsing_label(strategy_label)["super_main"]
        == str_mod.parsing_label(o["label"])["super_main"]
    ]


def combine_vars_to_get_future_spread_label(timestamp: int) -> str:
    """ """

    return f"futureSpread-open-{timestamp}"


def check_if_id_has_used_before(
    combined_result: str, id_checked: str, transaction_id: str
) -> bool:
    """
    id_checked: order_id, trade_id, label

    verifier: order_id or label?
    - order_id only one per order
    - one label could be processed couple of time (especially when closing the transactions)
    """

    id = f"{id_checked}"
    # log.error (f"id {id}")
    if combined_result != []:
        result_order_id = [o[id] for o in combined_result]

    label_is_exist: list = (
        False
        if (combined_result == [] or result_order_id == [])
        else False if transaction_id not in result_order_id else True
    )

    return label_is_exist


def provide_side_to_close_transaction(transaction: dict) -> str:
    """ """

    # determine side
    transaction_side = get_transaction_side(transaction)

    if transaction_side == "sell":
        side = "buy"

    if transaction_side == "buy":
        side = "sell"

    return side


def sum_order_under_closed_label_int(
    closed_orders_label_strategy: list, label_integer_open: int
) -> int:
    """ """

    if closed_orders_label_strategy:
        order_under_closed_label_int = [
            o for o in closed_orders_label_strategy if label_integer_open in o["label"]
        ]

    return (
        0
        if (closed_orders_label_strategy == [] or order_under_closed_label_int == [])
        else sum([o["amount"] for o in order_under_closed_label_int])
    )


def convert_list_to_dict(transaction: list) -> dict:

    # convert list to dict
    try:
        transaction = transaction[0]
    except:
        return transaction

    return transaction


def is_label_and_side_consistent(non_checked_strategies, params) -> bool:
    """ """

    # log.error (f"params {params}")
    label = get_transaction_label(params)

    is_consistent = True if "closed" in label else False
    # log.warning(f"params {params}")

    if bool([o for o in non_checked_strategies if (o in label)]):
        is_consistent = True

    else:

        if "open" in label:

            side = get_transaction_side(params)

            if side == "sell":

                is_consistent = (
                    True
                    if ("Short" in label or "hedging" in label or "custom" in label)
                    else False
                )

                log.warning(f"is_consistent {is_consistent}")

            if side == "buy":
                is_consistent = True if "Long" in label else False

    return is_consistent


def get_take_profit_pct(transaction: dict, strategy_config: dict) -> float:
    """ """

    try:
        tp_pct: float = transaction["profit_target_pct_transaction"]
    except:
        tp_pct: float = strategy_config["take_profit_pct"]

    return tp_pct


def reading_from_db(end_point, instrument: str = None, status: str = None) -> list:
    """ """
    from utilities import pickling, system_tools

    return pickling.read_data(
        system_tools.provide_path_for_file(end_point, instrument, status)
    )


def get_non_label_from_transaction(transactions) -> list:
    """ """

    return [] if transactions == [] else [o for o in transactions if o["label"] == ""]


def check_db_consistencies(
    instrument_name: str,
    trades_from_sqlite: list,
    positions_from_sub_account: list,
    order_from_sqlite_open: list,
    open_orders_from_sub_accounts: list,
) -> bool:
    """ """

    no_non_label_from_from_sqlite_open = (
        False if get_non_label_from_transaction(order_from_sqlite_open) != [] else True
    )

    len_from_sqlite_open = len(order_from_sqlite_open)

    len_open_orders_from_sub_accounts = len(open_orders_from_sub_accounts)
    #
    sum_my_trades_sqlite = (
        0
        if trades_from_sqlite == []
        else sum([o["amount"] for o in trades_from_sqlite])
    )

    size_from_position: int = (
        0
        if positions_from_sub_account == []
        else sum(
            [
                o["size"]
                for o in positions_from_sub_account
                if o["instrument_name"] == instrument_name
            ]
        )
    )

    log.error(
        f"size_is_consistent {sum_my_trades_sqlite == size_from_position} sum_my_trades_sqlite {sum_my_trades_sqlite} size_from_positions {size_from_position} "
    )
    return dict(
        trade_size_is_consistent=sum_my_trades_sqlite == size_from_position,
        order_is_consistent=(
            len_open_orders_from_sub_accounts == len_from_sqlite_open
            and no_non_label_from_from_sqlite_open
        ),
        no_non_label_from_from_sqlite_open=(
            False
            if get_non_label_from_transaction(order_from_sqlite_open) != []
            else True
        ),
    )


def get_basic_closing_paramaters(
    selected_transaction: list,
    closed_orders_label_strategy: list,
    closed_label_status: str = None,
) -> dict:
    """ """

    transaction: dict = convert_list_to_dict(selected_transaction)

    # provide dict placeholder for params
    params = {}

    # determine side
    side = provide_side_to_close_transaction(transaction)
    params.update({"side": side})

    basic_size = get_transaction_size(transaction)

    label_transaction = transaction["label"]
    label_integer_open = get_label_integer(label_transaction)

    sum_order_under_closed_label = sum_order_under_closed_label_int(
        closed_orders_label_strategy, label_integer_open
    )

    net_size = basic_size + sum_order_under_closed_label
    size_abs = provide_size_to_close_transaction(basic_size, net_size)

    size = size_abs * ensure_sign_consistency(side)

    closing_size_ok = check_if_next_closing_size_will_not_exceed_the_original(
        basic_size, net_size, size
    )

    log.debug(
        f"sum_order_under_closed_label {sum_order_under_closed_label} label_integer_open {label_integer_open}"
    )

    # size=exactly amount of transaction size
    params.update({"size": size if closing_size_ok else 0})

    if closed_label_status is None:

        label_closed: str = get_label("closed", label_transaction)

    else:

        label_closed: str = get_label(closed_label_status, label_transaction)

    params.update({"label": label_closed})

    params.update({"instrument_name": transaction["instrument_name"]})

    return params


def get_basic_closing_paramaters_combo_pair(
    selected_transactions: list,
) -> dict:
    """_summary_

    Args:
        selected_transactions (list): pair of instruments in the same futureSpread label

    Returns:
        dict: _description_
    """

    # provide dict placeholder for params
    params = {}

    # default type: limit
    params.update({"type": "limit"})

    # determine side
    side = "buy"
    params.update({"side": side})
    basic_size = abs(max([o["amount"] for o in selected_transactions]))

    label_integer_open = [o["label"] for o in selected_transactions][0]
    log.warning(f"label_integer_open {label_integer_open}")

    instrument_name = [o["combo_id"] for o in selected_transactions][0]

    params.update({"size": basic_size})

    label_closed: str = get_label("closed", label_integer_open)
    params.update({"label": label_closed})

    params.update({"instrument_name": instrument_name})

    return params


@dataclass(unsafe_hash=True, slots=True)
class ManageStrategy:
    """
    https://stackoverflow.com/questions/13646245/is-it-possible-to-make-abstract-classes-in-python/13646263#13646263

    """

    # @abstractmethod
    def understanding_the_market(self) -> None:
        """ """
        pass

    # @abstractmethod
    def risk_managament(self) -> None:
        """ """
        pass

    # @abstractmethod
    def opening_position(self) -> None:
        """ """
        pass

    # @abstractmethod
    def closing_position(self) -> None:
        """ """
        pass

    # @abstractmethod
    def cancelling_order(self) -> None:
        """ """
        pass

    # @abstractmethod
    def modifying_order(self) -> None:
        """ """
        pass


@dataclass(unsafe_hash=True, slots=True)
class BasicStrategy(ManageStrategy):
    """ """

    strategy_label: str
    strategy_parameters: dict

    def get_basic_opening_parameters(
        self,
        ask_price: float = None,
        bid_price: float = None,
        notional: float = None,
    ) -> dict:
        """ """

        # provide placeholder for params
        params = {}

        # default type: limit
        params.update({"type": "limit"})

        strategy_config: dict = self.strategy_parameters

        side: str = strategy_config["side"]

        params.update({"side": side})

        if side == "sell":
            params.update({"entry_price": ask_price})

        if side == "buy":
            params.update({"entry_price": bid_price})

        label_open: str = get_label("open", self.strategy_label)
        params.update({"label": label_open})

        return params

    def get_basic_closing_paramaters(
        self, selected_transaction: list, closed_orders_label_strategy: list
    ) -> dict:
        """ """

        return get_basic_closing_paramaters(
            selected_transaction, closed_orders_label_strategy
        )

    def get_basic_closing_paramaters_combo_pair(
        self, selected_transactions: list
    ) -> dict:
        """ """

        return get_basic_closing_paramaters_combo_pair(selected_transactions)



# # -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from dataclassy import dataclass, fields
from loguru import logger as log

from streaming_helper.strategies.deribit.basic_strategy import BasicStrategy

# user defined formula


def get_delta(my_trades_currency_strategy) -> int:
    """ """

    return (
        0
        if my_trades_currency_strategy == []
        else sum([o["amount"] for o in my_trades_currency_strategy])
    )


@dataclass(unsafe_hash=True, slots=True)
class FutureSpread(BasicStrategy):
    """ """

    orders_currency_strategy: list
    server_time: int
    my_trades_currency_strategy: list = None
    ticker_perpetual: dict = None
    delta: float = fields
    basic_params: object = fields

    def __post_init__(self):

        self.delta: float = get_delta(self.my_trades_currency_strategy)
        self.basic_params: str = BasicStrategy(
            self.strategy_label, self.strategy_parameters
        )

        log.critical(f"""delta  {self.delta} """)



# # -*- coding: utf-8 -*-

# built ins
import asyncio
from secrets import randbelow

# installed
from dataclassy import dataclass, fields
from loguru import logger as log

from streaming_helper.db_management.sqlite_management import update_status_data
from streaming_helper.strategies.deribit.basic_strategy import (
    BasicStrategy,
    are_size_and_order_appropriate,
    ensure_sign_consistency,
    is_label_and_side_consistent,
    is_minimum_waiting_time_has_passed,
    size_rounding,
)

from streaming_helper.restful_api.deribit import end_point_params_template
from streaming_helper.restful_api import connector
from streaming_helper.utilities.string_modification import parsing_label

# user defined formula


def get_transactions_len(result_strategy_label) -> int:
    """ """
    return 0 if result_strategy_label == [] else len([o for o in result_strategy_label])


def get_transactions_sum(result_strategy_label) -> int:
    """ """
    return (
        0
        if result_strategy_label == []
        else sum([o["amount"] for o in result_strategy_label])
    )


def get_label_integer(label: dict) -> bool:
    """ """

    return parsing_label(label)["int"]


def reading_from_pkl_data(end_point, currency, status: str = None) -> dict:
    """ """

    from utilities.pickling import read_data
    from utilities.system_tools import provide_path_for_file

    path: str = provide_path_for_file(end_point, currency, status)
    data = read_data(path)

    return data


def hedged_value_to_notional(notional: float, hedged_value: float) -> float:
    """ """
    return abs(hedged_value / notional)


def determine_opening_size(
    instrument_name: str,
    futures_instruments,
    side: str,
    max_position: float,
    factor: float,
) -> int:
    """ """
    sign = ensure_sign_consistency(side)

    proposed_size = max(1, int(abs(max_position) * factor))

    return size_rounding(instrument_name, futures_instruments, proposed_size) * sign


def get_waiting_time_factor(
    weighted_factor,
    strong_fluctuation: bool,
    some_fluctuation: bool,
) -> float:
    """
    Provide factor for size determination.
    """

    ONE_PCT = 1 / 100

    BEARISH_FACTOR = (
        weighted_factor["extreme"] * ONE_PCT
        if strong_fluctuation
        else weighted_factor["medium"] * ONE_PCT
    )

    return BEARISH_FACTOR if (strong_fluctuation or some_fluctuation) else ONE_PCT


def is_hedged_value_to_notional_exceed_threshold(
    notional: float, hedged_value: float, threshold: float
) -> float:
    """ """
    return hedged_value_to_notional(notional, hedged_value) > threshold


def max_order_stack_has_not_exceeded(len_orders: float, strong_market: float) -> bool:
    """ """
    if strong_market:
        max_order = True

    else:
        max_order = True if len_orders == 0 else False

    return max_order


def get_timing_factor(strong_bearish: bool, bearish: bool, threshold: float) -> float:
    """
    Determine order outstanding timing for size determination.
    strong bearish : 30% of normal interval
    bearish        : 6% of normal interval
    """

    ONE_PCT = 1 / 100

    bearish_interval_threshold = (
        (threshold * ONE_PCT * 30) if strong_bearish else (threshold * ONE_PCT * 60)
    )

    return (
        threshold * bearish_interval_threshold
        if (strong_bearish or bearish)
        else threshold
    )


def check_if_minimum_waiting_time_has_passed(
    strong_bullish: bool,
    bullish: bool,
    threshold: float,
    timestamp: int,
    server_time: int,
) -> bool:
    """ """

    cancel_allowed: bool = False

    time_interval = get_timing_factor(strong_bullish, bullish, threshold)

    minimum_waiting_time_has_passed: bool = is_minimum_waiting_time_has_passed(
        server_time, timestamp, threshold
    )

    # log.warning (f"minimum_waiting_time_has_passed {minimum_waiting_time_has_passed} time_interval {time_interval}")
    # log.warning (f"server_time {server_time} max_tstamp_orders {timestamp} threshold {threshold} { {server_time-timestamp} }")

    if minimum_waiting_time_has_passed:
        cancel_allowed: bool = True

    return cancel_allowed


def current_hedge_position_exceed_max_position(
    sum_my_trades_currency_str: int, max_position: float
) -> bool:
    """ """

    return abs(sum_my_trades_currency_str) > abs(max_position)


def net_size_of_label(my_trades_currency_strategy: list, transaction: dict) -> bool:
    """ """

    label_integer = get_label_integer(transaction["label"])

    return sum(
        [
            o["amount"]
            for o in my_trades_currency_strategy
            if str(label_integer) in o["label"]
        ]
    )


def net_size_not_over_bought(
    my_trades_currency_strategy: list, transaction: list
) -> bool:
    """ """
    return net_size_of_label(my_trades_currency_strategy, transaction) < 0


def size_multiply_factor(market_condition: dict) -> float:
    """ """

    strong_bearish = market_condition["strong_bearish"]

    bearish = market_condition["bearish"]
    weak_bullish = market_condition["weak_bullish"]
    weak_bearish = market_condition["weak_bearish"]
    bullish = market_condition["bullish"]
    strong_bullish = market_condition["strong_bullish"]

    multiply_factor = 1

    if bullish or strong_bullish:
        multiply_factor = 0
    else:
        if weak_bullish:
            multiply_factor = 0.8
        elif strong_bearish:
            multiply_factor = 1.5
        elif bearish:
            multiply_factor = 1.3
        elif weak_bearish:
            multiply_factor = 1.1

    return multiply_factor


def size_to_be_hedged(
    notional,
    sum_my_trades_currency_strategy,
    over_hedged_closing,
    market_condition: dict,
) -> float:
    """ """

    multiply = size_multiply_factor(market_condition)

    max_position = (
        (abs(notional) + sum_my_trades_currency_strategy) * -1
        if over_hedged_closing
        else notional
    )

    return multiply * max_position


async def modify_hedging_instrument(
    strong_bearish: bool,
    bearish: bool,
    instrument_attributes_futures_for_hedging: list,
    ticker_all: list,
    ticker_perpetual_instrument_name: dict,
    currency_upper: str,
) -> dict:

    if bearish or not strong_bearish:

        instrument_attributes_future = [
            o
            for o in instrument_attributes_futures_for_hedging
            if "PERPETUAL" not in o["instrument_name"]
            and currency_upper in o["instrument_name"]
        ]

        if len(instrument_attributes_future) > 1:
            index_attributes = randbelow(2)
            instrument_attributes = instrument_attributes_future[index_attributes]

        else:
            instrument_attributes = instrument_attributes_future[0]

        instrument_name: str = instrument_attributes["instrument_name"]

        instrument_ticker = [
            o for o in ticker_all if instrument_name in o["instrument_name"]
        ]

        # log.error (f"instrument_ticker {instrument_ticker}")
        if instrument_ticker:
            instrument_ticker = instrument_ticker[0]

        else:
                
            connection_url = end_point_params_template.basic_https()

            tickers_end_point= end_point_params_template.get_tickers_end_point(instrument_name)
            
            instrument_ticker = await connector.get_connected(
            connection_url,
            tickers_end_point)
            
        return instrument_ticker

    else:
        return ticker_perpetual_instrument_name


@dataclass(unsafe_hash=True, slots=True)
class HedgingSpot(BasicStrategy):
    """ """

    notional: float
    my_trades_currency_strategy: int
    market_condition: list
    index_price: float
    my_trades_currency_all: float
    sum_my_trades_currency_strategy: int = fields
    over_hedged_opening: bool = fields
    over_hedged_closing: bool = fields
    max_position: float = fields
    max_position: float = fields

    def __post_init__(self):

        self.sum_my_trades_currency_strategy = get_transactions_sum(
            self.my_trades_currency_strategy
        )
        self.over_hedged_opening = current_hedge_position_exceed_max_position(
            self.sum_my_trades_currency_strategy, self.notional
        )
        self.over_hedged_closing = self.sum_my_trades_currency_strategy > 0
        self.max_position = size_to_be_hedged(
            self.notional,
            self.sum_my_trades_currency_strategy,
            self.over_hedged_closing,
            self.market_condition,
        )

    #        log.info(
    #            f" max_position {self.max_position} over_hedged_opening {self.over_hedged_opening}  over_hedged_closing {self.over_hedged_closing} sum_my_trades_currency_strategy {self.sum_my_trades_currency_strategy}"
    #        )

    def get_basic_params(self) -> dict:
        """ """
        return BasicStrategy(self.strategy_label, self.strategy_parameters)

    async def understanding_the_market(self, threshold_market_condition) -> None:
        """ """

    async def risk_managament(self) -> None:
        """ """
        pass

    def opening_position(
        self,
        non_checked_strategies,
        instrument_name,
        futures_instruments,
        open_orders_label,
        params,
        SIZE_FACTOR,
        len_orders,
    ) -> bool:
        """ """

        order_allowed: bool = False

        max_position = self.max_position

        my_trades_currency_all_sum = get_transactions_sum(self.my_trades_currency_all)

        over_hendge_opening_all = my_trades_currency_all_sum < max_position

        log.debug(
            f"my_trades_currency_all_sum {my_trades_currency_all_sum} max_position {max_position} over_hendge_opening_all {over_hendge_opening_all}"
        )

        if len_orders == 0 and not over_hendge_opening_all:

            over_hedged_cls = self.over_hedged_closing

            size = determine_opening_size(
                instrument_name,
                futures_instruments,
                params["side"],
                self.max_position,
                SIZE_FACTOR,
            )

            sum_orders: int = get_transactions_sum(open_orders_label)

            size_and_order_appropriate_for_ordering: bool = (
                are_size_and_order_appropriate(
                    "add_position",
                    self.sum_my_trades_currency_strategy,
                    sum_orders,
                    size,
                    max_position,
                )
            )

            order_allowed: bool = (
                size_and_order_appropriate_for_ordering or over_hedged_cls
            )

            log.info(f"order_allowed {order_allowed} ")
            log.info(
                f" size_and_order_appropriate_for_ordering {size_and_order_appropriate_for_ordering} { (size_and_order_appropriate_for_ordering or over_hedged_cls)}"
            )

            if order_allowed:

                label_and_side_consistent = is_label_and_side_consistent(
                    non_checked_strategies, params
                )
                log.info(f"label_and_side_consistent {label_and_side_consistent} ")

                if label_and_side_consistent:  # and not order_has_sent_before:

                    params.update({"instrument_name": instrument_name})
                    params.update({"size": abs(size)})
                    params.update(
                        {"is_label_and_side_consistent": label_and_side_consistent}
                    )

                else:

                    order_allowed = False

        return order_allowed

    def closing_position(
        self,
        transaction: dict,
        exit_params: dict,
        bullish: bool,
        strong_bullish: bool,
        weak_bullish: bool,
        bid_price: float,
    ) -> bool:
        """ """

        order_allowed: bool = False

        # bid_price_is_lower = bid_price < transaction ["price"]
        over_hedged_closing = self.over_hedged_closing
        over_hedged_opening = self.over_hedged_opening

        if over_hedged_closing:

            order_allowed: bool = False

        else:

            size = exit_params["size"]

            log.debug(f"transaction {transaction}")
            log.warning(
                f"size != 0 {size} (bullish or strong_bullish) {(weak_bullish or bullish or strong_bullish)}"
            )

            if size != 0 and over_hedged_opening:

                if weak_bullish or bullish or strong_bullish:

                    exit_params.update({"entry_price": bid_price})

                    # convert size to positive sign
                    exit_params.update({"size": abs(size)})

                    order_allowed: bool = True

        log.warning(f"order_allowed {order_allowed} ")

        return order_allowed

    async def cancelling_orders(
        self,
        transaction: dict,
        orders_currency_strategy: list,
        server_time: int,
        strategy_params: list = None,
    ) -> bool:
        """ """

        cancel_allowed: bool = False

        ONE_SECOND = 1000
        ONE_MINUTE = ONE_SECOND * 60

        if strategy_params is None:
            hedging_attributes: dict = self.strategy_parameters
        else:
            hedging_attributes: dict = strategy_params

        waiting_minute_before_cancel = (
            hedging_attributes["waiting_minute_before_cancel"] * ONE_MINUTE
        )

        market_condition: dict = self.market_condition

        bullish, strong_bullish = (
            market_condition["bullish"],
            market_condition["strong_bullish"],
        )

        bearish, strong_bearish = (
            market_condition["bearish"],
            market_condition["strong_bearish"],
        )
        # neutral = market_condition["neutral_price"]

        try:
            timestamp: int = transaction["timestamp"]

        except:
            timestamp: int = transaction["last_update_timestamp"]

        if timestamp:

            if "open" in transaction["label"]:

                open_size_not_over_bought = self.over_hedged_opening

                open_orders_label: list = [
                    o for o in orders_currency_strategy if "open" in o["label"]
                ]

                len_open_orders: int = get_transactions_len(open_orders_label)

                log.info(
                    f"open_size_not_over_bought {open_size_not_over_bought} len_open_orders {len_open_orders}"
                )

                if not open_size_not_over_bought:

                    # Only one open order a time
                    if len_open_orders > 1:

                        cancel_allowed = True

                    else:

                        cancel_allowed: bool = check_if_minimum_waiting_time_has_passed(
                            strong_bearish,
                            bearish,
                            waiting_minute_before_cancel,
                            timestamp,
                            server_time,
                        )
                        log.info(f"cancel_allowed {cancel_allowed}")

                # cancel any orders when overhedged
                else:

                    if len_open_orders > 0:

                        cancel_allowed = True

            if "closed" in transaction["label"]:

                exit_size_not_over_bought = net_size_not_over_bought(
                    self.my_trades_currency_strategy, transaction
                )

                log.info(f"exit_size_not_over_bought {exit_size_not_over_bought}")
                if exit_size_not_over_bought:

                    closed_orders_label: list = [
                        o for o in orders_currency_strategy if "closed" in (o["label"])
                    ]

                    len_closed_orders: int = get_transactions_len(closed_orders_label)

                    log.info(f" len_closed_orders {len_closed_orders}")

                    if len_closed_orders > 1:

                        cancel_allowed = True

                    else:

                        cancel_allowed: bool = check_if_minimum_waiting_time_has_passed(
                            strong_bullish,
                            bullish,
                            waiting_minute_before_cancel,
                            timestamp,
                            server_time,
                        )
                        log.info(f" cancel_allowed {cancel_allowed}")

                # cancel any orders when overhedged
                else:

                    cancel_allowed = True

        return cancel_allowed

    async def modifying_order(self) -> None:
        """ """
        pass

    async def is_cancelling_orders_allowed(
        self,
        selected_transaction: list,
        orders_currency_strategy: list,
        server_time: int,
    ) -> dict:
        """ """

        cancel_allowed, cancel_id = False, None

        cancel_allowed = await self.cancelling_orders(
            selected_transaction, orders_currency_strategy, server_time
        )

        if cancel_allowed:
            cancel_id = selected_transaction["order_id"]

        return dict(cancel_allowed=cancel_allowed, cancel_id=cancel_id)

    async def is_send_open_order_allowed(
        self,
        non_checked_strategies,
        instrument_name: str,
        futures_instruments,
        orders_currency_strategy: list,
        ask_price: float,
        archive_db_table: str,
        trade_db_table: str,
    ) -> dict:
        """ """

        order_allowed = False

        open_orders_label: list = [
            o for o in orders_currency_strategy if "open" in o["label"]
        ]

        len_open_orders: int = get_transactions_len(open_orders_label)

        hedging_attributes = self.strategy_parameters

        market_condition = self.market_condition

        # bullish = market_condition["rising_price"]
        bearish = market_condition["bearish"]

        # strong_bullish = market_condition["strong_rising_price"]
        strong_bearish = market_condition["strong_bearish"]
        #        neutral = market_condition["neutral"]
        params: dict = self.get_basic_params().get_basic_opening_parameters(ask_price)

        weighted_factor = hedging_attributes["weighted_factor"]

        SIZE_FACTOR = get_waiting_time_factor(weighted_factor, strong_bearish, bearish)

        order_allowed: bool = self.opening_position(
            non_checked_strategies,
            instrument_name,
            futures_instruments,
            open_orders_label,
            params,
            SIZE_FACTOR,
            len_open_orders,
        )

        # additional test for possibilities of orphaned closed orders was existed
        if not order_allowed:

            if len_open_orders == 0:

                # orphaned closed orders: have closed only
                my_trades_orphan = [
                    o
                    for o in self.my_trades_currency_strategy
                    if "closed" in o["label"] and "open" not in (o["label"])
                ]

                if my_trades_orphan:

                    max_timestamp = max([o["timestamp"] for o in my_trades_orphan])

                    transaction = [
                        o for o in my_trades_orphan if max_timestamp == o["timestamp"]
                    ][0]

                    label_integer = get_label_integer(transaction["label"])

                    new_label = f"futureSpread-open-{label_integer}"

                    filter = "label"

                    await update_status_data(
                        archive_db_table,
                        "label",
                        filter,
                        transaction["label"],
                        new_label,
                        "=",
                    )

                    await update_status_data(
                        trade_db_table,
                        "label",
                        filter,
                        transaction["label"],
                        new_label,
                        "=",
                    )

        return dict(
            order_allowed=order_allowed and len_open_orders == 0,
            order_parameters=[] if not order_allowed else params,
        )

    async def is_send_exit_order_allowed(
        self,
        orders_currency_strategy_label_closed,
        bid_price: float,
        selected_transaction: list,
        # orders_currency_strategy: list
    ) -> dict:
        """ """
        order_allowed: bool = False

        transaction: dict = selected_transaction[0]

        exit_size_not_over_bought = net_size_not_over_bought(
            self.my_trades_currency_strategy, transaction
        )

        label_integer: int = get_label_integer(transaction["label"])

        closed_orders_int = [
            o
            for o in orders_currency_strategy_label_closed
            if str(label_integer) in o["label"]
        ]

        len_closed_orders_int = get_transactions_len(closed_orders_int)

        log.warning(f" len_closed_orders_int == 0 {len_closed_orders_int == 0}")
        log.error(f" closed_orders_int {closed_orders_int}")
        log.debug(f" label_integer {label_integer}")
        log.info(
            f" orders_currency_strategy_label_closed {orders_currency_strategy_label_closed}"
        )

        if len_closed_orders_int == 0 and exit_size_not_over_bought:

            market_condition = self.market_condition

            bullish, strong_bullish, weak_bullish = (
                market_condition["bullish"],
                market_condition["strong_bullish"],
                market_condition["weak_bullish"],
            )

            exit_params: dict = self.get_basic_params().get_basic_closing_paramaters(
                selected_transaction,
                orders_currency_strategy_label_closed,
            )

            log.warning(
                f"sum_my_trades_currency_strategy {self.sum_my_trades_currency_strategy} "
            )

            order_allowed = self.closing_position(
                transaction,
                exit_params,
                bullish,
                strong_bullish,
                weak_bullish,
                bid_price,
            )

            # default type: limit
            exit_params.update({"type": "limit"})

        return dict(
            order_allowed=order_allowed,
            order_parameters=([] if not order_allowed else exit_params),
        )

    async def send_contra_order_for_orphaned_closed_transctions(
        self,
        orders_currency_strategy_label_open,
        ask_price: float,
        selected_transaction: list,
    ) -> dict:
        """ """
        order_allowed: bool = False

        transaction: dict = selected_transaction[0]

        exit_size_not_over_bought = net_size_not_over_bought(
            self.my_trades_currency_strategy, transaction
        )

        label_integer: int = get_label_integer(transaction["label"])

        open_orders_int = [
            o
            for o in orders_currency_strategy_label_open
            if str(label_integer) in o["label"]
        ]

        len_open_orders_int = get_transactions_len(open_orders_int)

        log.warning(f" len_closed_orders_int == 0 {len_open_orders_int == 0}")

        if len_open_orders_int == 0 and exit_size_not_over_bought:

            exit_params: dict = self.get_basic_params().get_basic_closing_paramaters(
                selected_transaction,
                orders_currency_strategy_label_open,
                "contra",
            )

            log.warning(
                f"sum_my_trades_currency_strategy {self.sum_my_trades_currency_strategy} "
            )

            size = exit_params["size"]

            ask_price_is_higher = ask_price > transaction["price"]

            if size != 0 and ask_price_is_higher:

                exit_params.update({"entry_price": ask_price})

                # convert size to positive sign
                exit_params.update({"size": abs(size)})

                order_allowed: bool = True

            # default type: limit
            exit_params.update({"type": "limit"})

        return dict(
            order_allowed=order_allowed,
            order_parameters=([] if not order_allowed else exit_params),
        )



# -*- coding: utf-8 -*-

# built ins
import asyncio
from streaming_helper.utilities import error_handling

async def redis_channels(
    pubsub: object,
    redis_channels: list,
    purpose: str,
) -> None:
    """ """

    # get redis channels
    order_allowed_channel: str = redis_channels["order_is_allowed"]
    positions_update_channel: str = redis_channels["position_cache_updating"]
    ticker_cached_channel: str = redis_channels["ticker_cache_updating"]
    order_rest_channel: str = redis_channels["order_rest"]
    my_trade_receiving_channel: str = redis_channels["my_trade_receiving"]
    order_update_channel: str = redis_channels["order_cache_updating"]
    portfolio_channel: str = redis_channels["portfolio"]
    sqlite_updating_channel: str = redis_channels["sqlite_record_updating"]
    sub_account_cached_channel: str = redis_channels["sub_account_cache_updating"]
    market_analytics_channel: str = redis_channels["market_analytics_update"]
    my_trades_channel: str = redis_channels["my_trades_cache_updating"]

    try:

        match purpose:

            case "reconciling_size":
                channels = [
                    my_trade_receiving_channel,
                    order_allowed_channel,
                    portfolio_channel,
                    positions_update_channel,
                    sub_account_cached_channel,
                    ticker_cached_channel,
                ]
            case "processing_orders":
                channels = [
                    my_trade_receiving_channel,
                    portfolio_channel,
                    order_rest_channel,
                    order_update_channel,
                    sqlite_updating_channel,
                    sub_account_cached_channel,
                ]
            case "hedging_spot" | "future_spread":
                channels = [
                    market_analytics_channel,
                    order_update_channel,
                    ticker_cached_channel,
                    portfolio_channel,
                    my_trades_channel,
                    order_allowed_channel,
                    sub_account_cached_channel,
                ]
            case "cancelling_active_orders":
                channels = [
                    market_analytics_channel,
                    order_update_channel,
                    ticker_cached_channel,
                    portfolio_channel,
                    my_trades_channel,
                    sub_account_cached_channel,
                ]

        [await pubsub.subscribe(o) for o in channels]

    except Exception as error:

        error_handling.parse_error_message(error)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from loguru import logger as log

# user defined formula

from streaming_helper.restful_api.deribit import end_point_params_template
from streaming_helper.restful_api import connector
from streaming_helper.db_management import sqlite_management as db_mgt
from streaming_helper.channel_management import get_published_messages
from streaming_helper.channel_management.deribit import subscribing_to_channels
from streaming_helper.strategies.deribit.cash_carry import combo_auto as combo
from streaming_helper.strategies.deribit.hedging import hedging_spot

from streaming_helper.data_announcer.deribit import get_instrument_summary, starter
from streaming_helper.utilities import (
    string_modification as str_mod,
    error_handling,
)


async def cancelling_orders(
    client_id,
    client_secret,
    currencies: list,
    client_redis: object,
    config_app: list,
    initial_data_subaccount: dict,
    redis_channels: list,
    strategy_attributes: list,
) -> None:
    """ """

    try:

        # connecting to redis pubsub
        pubsub: object = client_redis.pubsub()

        # subscribe to channels
        await subscribing_to_channels.redis_channels(
            pubsub,
            redis_channels,
            "cancelling_active_orders",
        )

        relevant_tables = config_app["relevant_tables"][0]

        order_db_table = relevant_tables["orders_table"]

        strategy_attributes_active = [
            o for o in strategy_attributes if o["is_active"] == True
        ]

        active_strategies = [o["strategy_label"] for o in strategy_attributes_active]

        relevant_tables = config_app["relevant_tables"][0]

        order_db_table = relevant_tables["orders_table"]

        settlement_periods = get_settlement_period(strategy_attributes)

        futures_instruments = await get_instrument_summary.get_futures_instruments(
            currencies,
            settlement_periods,
        )

        instrument_attributes_futures_all = futures_instruments["active_futures"]

        # get redis channels
        market_analytics_channel: str = redis_channels["market_analytics_update"]
        ticker_cached_channel: str = redis_channels["ticker_cache_updating"]
        portfolio_channel: str = redis_channels["portfolio"]
        my_trades_channel: str = redis_channels["my_trades_cache_updating"]
        order_update_channel: str = redis_channels["order_cache_updating"]
        sub_account_cached_channel: str = redis_channels["sub_account_cache_updating"]

        await asyncio.sleep(5)

        cached_ticker_all = []

        not_cancel = True

        sub_account_cached_params = initial_data_subaccount["params"]

        sub_account_cached = sub_account_cached_params["data"]

        cached_orders = sub_account_cached["orders_cached"]

        query_trades = f"SELECT * FROM  v_trading_all_active"

        result_template = str_mod.message_template()

        my_trades_active_from_db = await db_mgt.executing_query_with_return(
            query_trades
        )

        initial_data_my_trades_active = starter.my_trades_active_combining(
            my_trades_active_from_db,
            my_trades_channel,
            result_template,
        )

        my_trades_active_all = initial_data_my_trades_active["params"]["data"]
        
        connection_url = end_point_params_template.basic_https()

        # get portfolio from exchg
        subaccounts_end_point= end_point_params_template.get_subaccounts_end_point()
        
        subaccounts_params = end_point_params_template.get_subaccounts_params()
        
        portfolio_from_exchg = await connector.get_connected(
            connection_url,
            subaccounts_end_point,
            client_id,
            client_secret,
            subaccounts_params,
            )
            
        initial_data_portfolio = starter.portfolio_combining(
            portfolio_from_exchg,
            portfolio_channel,
            result_template,
        )

        portfolio_all = initial_data_portfolio["params"]["data"]

        market_condition_all = []

        while not_cancel:

            try:

                message_byte = await pubsub.get_message()

                params = await get_published_messages.get_redis_message(message_byte)

                data = params["data"]

                message_channel = params["channel"]

                if market_analytics_channel in message_channel:

                    market_condition_all = data

                if portfolio_channel in message_channel:

                    portfolio_all = data["cached_portfolio"]

                if sub_account_cached_channel in message_channel:

                    cached_orders = data["open_orders"]

                    my_trades_active_all = data["my_trades"]

                if my_trades_channel in message_channel:

                    my_trades_active_all = data

                if order_update_channel in message_channel:

                    cached_orders = data["open_orders"]

                if ticker_cached_channel in message_channel and market_condition_all:

                    cached_ticker_all = data["data"]

                    server_time = data["server_time"]

                    currency_upper = data["currency_upper"]

                    currency_lower = currency_upper.lower()

                    instrument_name_perpetual = f"{currency_upper}-PERPETUAL"

                    market_condition = [
                        o
                        for o in market_condition_all
                        if instrument_name_perpetual in o["instrument_name"]
                    ][0]

                    portfolio = [
                        o
                        for o in portfolio_all
                        if currency_upper == o["currency"].upper()
                    ][0]
                    #! need to update currency to upper from starter

                    equity: float = portfolio["equity"]

                    ticker_perpetual_instrument_name = [
                        o
                        for o in cached_ticker_all
                        if instrument_name_perpetual in o["instrument_name"]
                    ][0]

                    index_price = get_index(ticker_perpetual_instrument_name)

                    my_trades_currency_all_transactions: list = (
                        []
                        if not my_trades_active_all
                        else [
                            o
                            for o in my_trades_active_all
                            if currency_upper in o["instrument_name"]
                        ]
                    )

                    my_trades_currency_all: list = (
                        []
                        if my_trades_currency_all_transactions == 0
                        else [
                            o
                            for o in my_trades_currency_all_transactions
                            if o["instrument_name"]
                            in [
                                o["instrument_name"]
                                for o in instrument_attributes_futures_all
                            ]
                        ]
                    )

                    orders_currency = (
                        []
                        if not cached_orders
                        else [
                            o
                            for o in cached_orders
                            if currency_upper in o["instrument_name"]
                        ]
                    )

                    if index_price is not None and equity > 0:

                        my_trades_currency: list = [
                            o for o in my_trades_currency_all if o["label"] is not None
                        ]

                        notional: float = compute_notional_value(index_price, equity)

                        for strategy in active_strategies:

                            strategy_params = [
                                o
                                for o in strategy_attributes
                                if o["strategy_label"] == strategy
                            ][0]

                            my_trades_currency_strategy = [
                                o
                                for o in my_trades_currency
                                if strategy in (o["label"])
                            ]

                            orders_currency_strategy = (
                                []
                                if not orders_currency
                                else [
                                    o
                                    for o in orders_currency
                                    if strategy in (o["label"])
                                ]
                            )

                            await cancelling_double_ids(
                                client_redis,      
                                client_id,
                                client_secret,
                                order_db_table,
                                orders_currency_strategy,
                            )

                            if "futureSpread" in strategy:

                                strategy_params = [
                                    o
                                    for o in strategy_attributes
                                    if o["strategy_label"] == strategy
                                ][0]

                                combo_auto = combo.ComboAuto(
                                    strategy,
                                    strategy_params,
                                    orders_currency_strategy,
                                    server_time,
                                    market_condition,
                                    my_trades_currency_strategy,
                                )

                                if orders_currency_strategy:
                                    for order in orders_currency_strategy:
                                        cancel_allowed: dict = (
                                            await combo_auto.is_cancelling_orders_allowed(
                                                order,
                                                server_time,
                                            )
                                        )

                                        if cancel_allowed["cancel_allowed"]:
                                            await if_cancel_is_true(
                                                client_id,
                                                client_secret,
                                                order_db_table,
                                                cancel_allowed,
                                            )

                                            not_cancel = False

                            if "hedgingSpot" in strategy:

                                max_position: int = notional * -1

                                hedging = hedging_spot.HedgingSpot(
                                    strategy,
                                    strategy_params,
                                    max_position,
                                    my_trades_currency_strategy,
                                    market_condition,
                                    index_price,
                                    my_trades_currency_all,
                                )

                                if orders_currency_strategy:

                                    for order in orders_currency_strategy:
                                        cancel_allowed: dict = (
                                            await hedging.is_cancelling_orders_allowed(
                                                order,
                                                orders_currency_strategy,
                                                server_time,
                                            )
                                        )

                                        if cancel_allowed["cancel_allowed"]:
                                            await if_cancel_is_true(
                                                client_id,
                                                client_secret,
                                                order_db_table,
                                                cancel_allowed,
                                            )

                                            not_cancel = False

            except Exception as error:

                await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
            )

                continue

            finally:
                await asyncio.sleep(0.001)

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
            )


def get_settlement_period(strategy_attributes) -> list:

    return str_mod.remove_redundant_elements(
        str_mod.remove_double_brackets_in_list(
            [o["settlement_period"] for o in strategy_attributes]
        )
    )


def compute_notional_value(
    index_price: float,
    equity: float,
) -> float:
    """ """
    return index_price * equity


def get_index(ticker: dict) -> float:

    try:

        index_price = ticker["index_price"]

    except:

        index_price = []

    if index_price == []:
        index_price = ticker["estimated_delivery_price"]

    return index_price


async def cancel_the_cancellables(
    client_id: str,
    client_secret: str,
    order_db_table: str,
    currency: str,
    cancellable_strategies: list,
    open_orders_sqlite: list = None,
) -> None:

    log.critical(f" cancel_the_cancellables {currency}")

    where_filter = f"order_id"

    column_list = "label", where_filter

    if open_orders_sqlite is None:
        open_orders_sqlite: list = (
            await db_mgt.executing_query_based_on_currency_or_instrument_and_strategy(
                "orders_all_json", currency.upper(), "all", "all", column_list
            )
        )

    if open_orders_sqlite:

        for strategy in cancellable_strategies:
            open_orders_cancellables = [
                o for o in open_orders_sqlite if strategy in o["label"]
            ]

            if open_orders_cancellables:
                open_orders_cancellables_id = [
                    o["order_id"] for o in open_orders_cancellables
                ]

                for order_id in open_orders_cancellables_id:

                    await cancel_by_order_id(
                        client_id,
                        client_secret,
                        order_db_table,
                        order_id,
                    )


async def cancel_by_order_id(
    client_id: str,
    client_secret: str,
    order_db_table: str,
    open_order_id: str,
) -> None:

    where_filter = f"order_id"

    await db_mgt.deleting_row(
        order_db_table,
        "databases/trading.sqlite3",
        where_filter,
        "=",
        open_order_id,
    )
    
    connection_url = end_point_params_template.basic_https()

    cancel_end_point= end_point_params_template.cancel_order()
    
    cancel_params = end_point_params_template.get_cancel_order_params(open_order_id)
    
    result = await connector.get_connected(
        connection_url,
        cancel_end_point,
        client_id,
        client_secret,
        cancel_params)
        

    try:
        if (result["error"]["message"]) == "not_open_order":
            log.critical(f"CANCEL non-existing order_id {result} {open_order_id}")

    except:

        log.critical(f"""CANCEL_by_order_id {result["result"]} {open_order_id}""")

        return result


async def if_cancel_is_true(
    client_id: str,
    client_secret: str,
    order_db_table: str,
    order: dict,
) -> None:
    """ """

    if order["cancel_allowed"]:

        # get parameter orders
        await cancel_by_order_id(
            client_id,
            client_secret, 
            order_db_table,
            order["cancel_id"],
        )


async def cancelling_double_ids(
    client_id,
    client_secret,
    client_redis,
    order_db_table: str,
    orders_currency_strategy: list,
) -> None:
    """ """

    if orders_currency_strategy:

        outstanding_order_id: list = str_mod.remove_redundant_elements(
            [o["label"] for o in orders_currency_strategy]
        )

        for label in outstanding_order_id:

            orders = [o for o in orders_currency_strategy if label in o["label"]]

            len_label = len(orders)

            if len_label > 1:

                for order in orders:
                    log.critical(f"double ids {label}")
                    log.critical(orders)

                    await cancel_by_order_id(
                        client_id,
                        client_secret,
                        order_db_table,
                        order["order_id"],
                    )

                    await error_handling.parse_error_message_with_redis(
            client_redis,
            f"avoiding double ids - {orders}",
            
            )
